/* (C) 2011 by Max Suraev
 * (C) 2010 by Holger Hans Peter Freyther
 * (C) 2010 by Harald Welte <laforge@gnumonks.org>
 * (C) 2010 by Alex Badea <vamposdecampos@gmail.com>
 *
 * All Rights Reserved
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <signal.h>
#include <stdlib.h>
#include <time.h>
#include <getopt.h>
#include <unistd.h>
#include <osmocom/core/signal.h>
#include <osmocom/core/msgb.h>
#include <osmocom/core/talloc.h>
#include <osmocom/core/utils.h>
#include <osmocom/gsm/rsl.h>
#include <osmocom/gsm/tlv.h>
#include <osmocom/gsm/gsm48_ie.h>
#include <osmocom/gsm/gsm48.h>
#include <osmocom/gsm/protocol/gsm_04_08.h>
#include <osmocom/bb/common/logging.h>
#include <osmocom/bb/common/lapdm.h>
#include <osmocom/bb/common/networks.h>
#include <osmocom/bb/common/osmocom_data.h>
#include <osmocom/bb/common/l1ctl.h>
#include <osmocom/bb/common/l23_app.h>
#include <osmocom/bb/misc/rslms.h>
#include <osmocom/bb/misc/layer3.h>

extern struct log_target * stderr_target;
extern void * l23_ctx; //The maximum possible number of RACH slots with a single-timeslot CCCH is 200.
size_t rach_count, rach_max = 200;
int has_si1, rach_cb_not_installed, report_rx, flip, target_report, ccch_mode;
struct timer_list rach_timer;
uint8_t ra;
uint16_t rach_offset;
uint16_t offsets[] = {3,4,13,35,44,45};//rach slot+1
size_t offset_counter = 0;
static struct gsm48_sysinfo sysinfo;

char * logname = "/var/tmp/rach.log";

inline char * ccch_mode_print(int mode)
{
    switch(mode)
    {
	case CCCH_MODE_NONE: return "CCCH_MODE_NONE";
	case CCCH_MODE_COMBINED: return "CCCH_MODE_COMBINED";
	case CCCH_MODE_NON_COMBINED: return "CCCH_MODE_NON_COMBINED";
	default: return "unknown CCCH mode";
    }
}

inline int si_fail(void * hdr, void * buf, size_t len, char * msg)
{
    if(!memcmp(hdr, buf, len)) { LOGP(DRR, LOGL_ERROR, "%s memory check failed\n", msg); return 1; }
    return 0;
}

inline static void dump_bcch(struct osmocom_ms * ms, struct msgb * msg)
{
    struct gsm48_system_information_type_header * si_hdr;
    struct gsm48_sysinfo *s = &sysinfo;
    struct abis_rsl_cchan_hdr * r_cch;

    si_hdr = msgb_l3(msg);
    struct gsm_sysinfo_freq cell_arfcns[1024];
    bzero(&cell_arfcns, sizeof(cell_arfcns));

/* GSM 05.02 §6.3.1.3 Mapping of BCCH data */
    switch(si_hdr->system_information)
    {
	case GSM48_MT_RR_SYSINFO_1:
	    if(si_fail(si_hdr, s->si1_msg, sizeof(s->si1_msg), "SI1")) return;
	    if(!has_si1)
	    {
		gsm48_decode_sysinfo1(s, (struct gsm48_system_information_type_1 *) si_hdr, msgb_l3len(msg));
		struct gsm48_system_information_type_1 * si1 = (struct gsm48_system_information_type_1 *)  msgb_l3(msg);
		gsm48_decode_freq_list(cell_arfcns, si1->cell_channel_description, sizeof(si1->cell_channel_description), 0xff, 0x01);
		has_si1 = 1;
		LOGP(DRR, LOGL_ERROR, "SI1 received.\n");
	    }
	break;
	case GSM48_MT_RR_SYSINFO_2:
	    if(si_fail(si_hdr, s->si2_msg, sizeof(s->si2_msg), "SI2")) return;
	    gsm48_decode_sysinfo2(s, (struct gsm48_system_information_type_2 *) si_hdr, msgb_l3len(msg));
	break;
	case GSM48_MT_RR_SYSINFO_2bis:
	    if(si_fail(si_hdr, s->si2b_msg, sizeof(s->si2b_msg), "SI2bis")) return;
	    gsm48_decode_sysinfo2bis(s, (struct gsm48_system_information_type_2bis *) si_hdr, msgb_l3len(msg));
	break;
	case GSM48_MT_RR_SYSINFO_2ter:
	    if(si_fail(si_hdr, s->si2t_msg, sizeof(s->si2t_msg), "SI2ter")) return;
	    gsm48_decode_sysinfo2ter(s, (struct gsm48_system_information_type_2ter *) si_hdr, msgb_l3len(msg));
	break;
	case GSM48_MT_RR_SYSINFO_3:
	    if(si_fail(si_hdr, s->si3_msg, sizeof(s->si3_msg), "SI3")) return;
	    if(ccch_mode == CCCH_MODE_NONE)
	    {
		gsm48_decode_sysinfo3(s, (struct gsm48_system_information_type_3 *) si_hdr, msgb_l3len(msg));
		struct gsm48_system_information_type_3 * si3 = (struct gsm48_system_information_type_3 *) msgb_l3(msg);
		//ccch_mode = (s->ccch_conf == 1) ? CCCH_MODE_COMBINED : CCCH_MODE_NON_COMBINED;
		if(si3->control_channel_desc.ccch_conf == RSL_BCCH_CCCH_CONF_1_C) ccch_mode = CCCH_MODE_COMBINED;
		else ccch_mode = CCCH_MODE_NON_COMBINED;
		r_cch = msgb_l2(msg);
		LOGP(DRR, LOGL_ERROR, "CCCH mode set to %s: %d\n", ccch_mode_print(ccch_mode), l1ctl_tx_ccch_mode_req(ms, ccch_mode));
	    }
	break;
	case GSM48_MT_RR_SYSINFO_4:
	    if(si_fail(si_hdr, s->si4_msg, sizeof(s->si4_msg), "SI4")) return;
	    gsm48_decode_sysinfo4(s, (struct gsm48_system_information_type_4 *) si_hdr, msgb_l3len(msg));
	break;
	case GSM48_MT_RR_SYSINFO_5: break;
	case GSM48_MT_RR_SYSINFO_5bis: break;
	case GSM48_MT_RR_SYSINFO_5ter: break;
	case GSM48_MT_RR_SYSINFO_6: break;
	case GSM48_MT_RR_SYSINFO_7: break;
	case GSM48_MT_RR_SYSINFO_8: break;
	case GSM48_MT_RR_SYSINFO_9: break;
	case GSM48_MT_RR_SYSINFO_13: break;
	case GSM48_MT_RR_SYSINFO_16: break;
	case GSM48_MT_RR_SYSINFO_17: break;
	default: LOGP(DRR, LOGL_ERROR, "\tUnknown SI\n"); break;
    };
}

uint8_t next_ra(uint8_t r)
{	
	r++;
	struct gsm48_sysinfo *s = &sysinfo;
	
	if(s->neci)
	{//NECI, New Establishment Causes - supported very early assignment.
		if(r < 64) return 64;
		if(r < 80) return r;
	}
	else
	{
		if(r < 224) return 224;
		return r;
	}

	return 0;
}

static void rach_send_cb(void * data)
{
    struct osmocom_ms * ms = data;
    struct gsm48_sysinfo *s = &sysinfo;
    uint16_t offset = 0;

    if(rach_count == rach_max)
    {
	LOGP(DRR, LOGL_ERROR, "MAX RACH limit %zu reached: terminating...\n", rach_count);
	exit(0);
    }

    if(has_si1 && ccch_mode != CCCH_MODE_NONE)
    {
	if(target_report)
	{
	    LOGP(DRR, LOGL_ERROR, "target locked on ARFCN=%d, NECI=%d: MCC=%s MNC=%s (%s, %s)\n", ms->test_arfcn, s->neci, gsm_print_mcc(s->mcc), gsm_print_mnc(s->mnc), gsm_get_mcc(s->mcc), gsm_get_mnc(s->mcc, s->mnc));
	    target_report = 0;
	}
// prepare RACH parameters
	uint8_t chan_req_val, chan_req_mask;
	chan_req_mask = (s->neci) ? 0x0f : 0x1f;
	chan_req_val = (s->neci) ? 0x01 : 0xe0;	
	ra = next_ra(ra);
	offset = 32; // dense enough for testing

	rach_count++;
	l1ctl_tx_rach_req(ms, ra, offset, ccch_mode);
    }
    bsc_schedule_timer(&rach_timer, 0, 100); // reschedule ourselves in 1 second, 0 milliseconds
}

int gsm48_rx_ccch(struct msgb *msg, struct osmocom_ms *ms) { return 0; } // make linker happy

int gsm48_rx_bcch(struct msgb *msg, struct osmocom_ms *ms)
{
    if(has_si1 && ccch_mode != CCCH_MODE_NONE) return 0;
    if(msgb_l3len(msg) != 23) { LOGP(DRR, LOGL_NOTICE, "Invalid BCCH message length\n"); return -EINVAL; }

    dump_bcch(ms, msg);

    if(report_rx)
    {
        LOGP(DRR, LOGL_ERROR, "obtained BCCH, dumping...\n");
	struct rx_meas_stat * rm = &ms->meas;
	LOGP(DRR, LOGL_ERROR, "RX level: %d\n", rm->rxlev / rm->frames - 110);
	report_rx = 0;
    }

    if(rach_cb_not_installed)
    {
	rach_timer.cb = &rach_send_cb;
	rach_timer.data = ms;
	LOGP(DRR, LOGL_ERROR, "scheduling RACH callback...\n");
	bsc_schedule_timer(&rach_timer, 1, 0); // schedule rach sender
	rach_cb_not_installed = 0;
    }
    return 0;
}

inline char * s_l1_enum_print(unsigned int signal)
{
    switch(signal)
    {
	case S_L1CTL_RESET: return "S_L1CTL_RESET";
	case S_L1CTL_FBSB_RESP: return "S_L1CTL_FBSB_RESP";
	case S_L1CTL_PM_RES: return "S_L1CTL_PM_RES";
	case S_L1CTL_PM_DONE: return "S_L1CTL_PM_DONE";
	case S_L1CTL_CCCH_MODE_CONF: return "S_L1CTL_CCCH_MODE_CONF";
	case S_L1CTL_TCH_MODE_CONF: return "S_L1CTL_TCH_MODE_CONF";
	case S_L1CTL_LOSS_IND: return "S_L1CTL_LOSS_IND";
	case S_L1CTL_FBSB_ERR: return "S_L1CTL_FBSB_ERR";
	default: return "unknown L1 value";
    }
}

void layer3_app_reset(void)
{/* Reset state */
    srand(time(NULL));
    has_si1 = 0;
    ccch_mode = CCCH_MODE_NONE;
    rach_count = 0;
    report_rx = 1;
    flip = 1;
    ra = 0;
    rach_offset = 0;
    target_report = 1;
    rach_cb_not_installed = 1;
}

static int l23_cfg_print_help()
{
    printf("\nApplication specific\n");
    printf("  -l --logfile LOGFILE	Logfile for the cell log.\n");
    printf("  -r --rach RACH	Number of RACH bursts to send.\n");
    return 0;
}

static int l23_cfg_handle(int c, const char *optarg)
{
    switch(c)
    {
	case 'l': logname = talloc_strdup(l23_ctx, optarg); break;
	case 'r': rach_max = atoi(optarg); break;
    }
    return 0;
}

static int signal_cb(unsigned int subsys, unsigned int signal, void * handler_data, void * signal_data)
{
    struct osmocom_ms * ms;
    struct osmobb_fbsb_res * fbsbr;

    if(subsys != SS_L1CTL) return 0;

    switch(signal)
    {
	case S_L1CTL_RESET:
	    ms = signal_data;
	    LOGP(DRR, LOGL_ERROR, "requesting FBSB from L1...\n");
	    layer3_app_reset();
	    return l1ctl_tx_fbsb_req(ms, ms->test_arfcn, L1CTL_FBSB_F_FB01SB, 100, 0, CCCH_MODE_NONE);
	break;
	case S_L1CTL_FBSB_ERR:
	    ms = signal_data;
	    LOGP(DRR, LOGL_ERROR, "FBSB error, resetting L1...\n");
	    l1ctl_tx_reset_req(ms, L1CTL_RES_T_FULL);
	break;
	case S_L1CTL_FBSB_RESP:
	    fbsbr = signal_data;
	    LOGP(DRR, LOGL_ERROR, "FBSB responce: BSIC %d, %d\n", fbsbr->bsic >> 3, fbsbr->bsic & 7);
	break;
	default: LOGP(DRR, LOGL_ERROR, "unhandled callback %u <%s> fired...\n", signal, s_l1_enum_print(signal));
    }
    return 0;
}

static int l23_cfg_supported() { return L23_OPT_TAP | L23_OPT_DBG; }

static int l23_getopt_options(struct option **options)
{
    static struct option opts [] = {
	{"logfile", 1, 0, 'l'},
	{"rach", 1, 0, 'r'},
    };

    *options = opts;
    return ARRAY_SIZE(opts);
}

int l23_app_init(struct osmocom_ms *ms)
{
    srand(time(NULL));

    LOGP(DRR, LOGL_ERROR, "MAX RACH is %zu\n", rach_max);
    LOGP(DRR, LOGL_ERROR, "registering signal handler...\n");
    register_signal_handler(SS_L1CTL, &signal_cb, NULL); //L1 signal handlers
    LOGP(DRR, LOGL_ERROR, "resetting L1...\n");
    l1ctl_tx_reset_req(ms, L1CTL_RES_T_FULL);
    LOGP(DRR, LOGL_ERROR, "L3 init...\n");
    return layer3_init(ms);
}

static struct l23_app_info info = {
    .copyright	= "Copyleft (C) 2011 Max Suraev\n",
    .contribution	= "Based on code by Harald Welte <laforge@gnumonks.org> with contributions from Holger Hans Peter Freyther\n",
    .getopt_string	= "g:p:l:r:nf:b:",
    .cfg_supported	= l23_cfg_supported,
    .cfg_getopt_opt = l23_getopt_options,
    .cfg_handle_opt	= l23_cfg_handle,
    .cfg_print_help	= l23_cfg_print_help,
};

struct l23_app_info *l23_app_info() { return &info; }
